import types
import sys
from library_class import LibraryClass
from decorules.decorators import raise_if_false_on_class, raise_if_false_on_instance, enforces_instance_rules

is_multiplier_larger_than_1 = lambda x: x.MULTIPLIER > 1.0
smallest_coordinate_times_multiplier_larger_than_largest = lambda x: min(x.coordinates) * x.MULTIPLIER > max(
    x.coordinates)


@raise_if_false_on_instance(smallest_coordinate_times_multiplier_larger_than_largest,
                            ValueError,
                            "Checks that the multiplier times the smallest coordinate is larger than the largest "
                            "coordinate")
@raise_if_false_on_class(is_multiplier_larger_than_1,
                         AttributeError,
                         "Checks that the multiplier is larger than 2.0")
class ClientClass(LibraryClass):
    MULTIPLIER = 2.5

    @enforces_instance_rules
    def append(self, value=0.0):
        self.coordinates.append(value)


class SomeBaseClass:
    def foo(self):
        print(f"Executing method {sys._getframe().f_code.co_name}")
        pass


class MultipleInheritanceClass(SomeBaseClass, ClientClass):
    @enforces_instance_rules
    def multiply_multiplier(self, value=1.0):
        self.MULTIPLIER = self.MULTIPLIER * value


class SomeOtherClass(SomeBaseClass):
    @enforces_instance_rules
    def foo(self):
        print(f"Executing method {sys._getframe().f_code.co_name}")
        pass


class LayerClass1(ClientClass):
    pass


class LayerClass2(LayerClass1):
    pass


class LayerClass3(LayerClass2):
    pass


class LayerClass4(LayerClass3):
    pass


def main():
    a = LibraryClass("a")
    a.testme()
    b = LibraryClass("b", 0.2, 0.5)
    b.testme()
    try:
        c = LibraryClass("c", -2.0, 0.5)
        c.testme()
    except ValueError as ve:
        print(ve)
        pass
    d = ClientClass("d", 0.1, 0.1, 0.05)
    d.testme()
    try:
        e = ClientClass("e", 0.1, 0.1, 0.025)
        e.testme()
    except ValueError as ve:
        print(ve)
        pass
    try:
        f = ClientClass("f", 0.8, 0.8, 0.4)
        f.testme()
    except ValueError as ve:
        print(ve)
        pass
    g = ClientClass("g", 0.1, 0.1, 0.05)
    g.testme()
    g.append(0.06)  # should not break an instance rule
    g.testme()
    try:
        g.append(max(g.coordinates) * ClientClass.MULTIPLIER)  # should break
        g.testme()
    except ValueError as ve:
        print(ve)
        pass
    h = MultipleInheritanceClass("h", 0.1, 0.1, 0.05)
    h.multiply_multiplier(1.0)
    h.testme()
    h.multiply_multiplier(10.0)
    h.testme()
    try:
        h.multiply_multiplier(0.001)
        h.testme()
    except ValueError as ve:
        print(ve)
        pass
    i = SomeOtherClass()
    try:
        i.foo()
    except TypeError as te:
        print(te)
        pass

    j = LayerClass4("j", 0.1, 0.1, 0.05)
    j.append(0.06)  # should not break an instance rule#
    try:
        j.append(max(j.coordinates) * LayerClass4.MULTIPLIER)  # should break
    except ValueError as ve:
        print(ve)
        pass


if __name__ == "__main__":
    main()



from functools import wraps, partial
import operator


def member_enforcer(enforced_key: str,
                    enforced_type: type,
                    comparison_value=None,
                    operator_used=operator.eq,
                    attrs_used: dict = None):
    """
    member_enforcer

    The member_enforcer function is a factory that creates a custom enforcement function to validate attributes of objects or classes. It checks if an attribute (specified by enforced_key) exists and meets certain conditions, including type checking and comparison operations.
    Parameters

    :param enforced_key (str):The name of the attribute (member) to be checked on the instance or class.
    :param enforced_type (type): The expected type of the attribute. The attribute's type must be a subclass of this type for the check to pass.
    :param comparison_value (optional, default: None): A value to compare the attribute against. If provided, the comparison is performed using the operator_used function.
    :param operator_used (optional, default: operator.eq): A function used to compare the attribute's value with comparison_value. Defaults to equality (operator.eq). Other examples include operator.lt for less than, operator.gt for greater than, etc.
    :param attrs_used (optional, default: None): A dictionary providing fallback attributes. If the primary attribute (enforced_key) is not found on the object, the function will look it up in this dictionary. Used for checking the static attributes of a class
    :return: bool
    """

    def key_type_comparison_enforcer(instance_or_type,
                                     enforced_type: type,
                                     enforced_key: str,
                                     comparison_value=comparison_value,
                                     operator_used=operator_used,
                                     attrs_used=attrs_used):
        member_object = getattr(instance_or_type, enforced_key, None)
        if member_object is None:
            if attrs_used is not None:
                member_object = attrs_used.get(enforced_key, None)
        if member_object is None:
            return False
        else:
            if issubclass(type(member_object), enforced_type):
                if comparison_value is not None and operator_used is not None:
                    return operator_used(member_object, comparison_value)
                else:
                    return True
            else:
                return False
        pass

    return partial(key_type_comparison_enforcer,
                   enforced_type=enforced_type,
                   enforced_key=enforced_key,
                   comparison_value=comparison_value,
                   operator_used=operator_used,
                   attrs_used=attrs_used)


def false_on_raise_else_true(func):
    # will be used when we 'transfer' enforced rules
    @wraps(func)
    def func_wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
            return True
        except Exception as ex:
            return False

    return func_wrapper



import types
from functools import partial
from decorules.has_enforced_rules import HasEnforcedRules
from decorules.decorators import raise_if_false_on_class, raise_if_false_on_instance
from decorules.predicates import key_type_enforcer
from decorules.utils import member_enforcer
from collections.abc import Iterable


def key_structure_enforcer(input_type,
                           enforced_type: type,
                           enforced_key: str):
    member_object = getattr(input_type, enforced_key, None)
    if member_object is None:
        return False
    else:
        return issubclass(type(member_object), enforced_type)
    pass


are_coordinates_within_distance_1 = lambda y: (sum([x ** 2 for x in y.coordinates]) ** 0.5) <= 1.0


@raise_if_false_on_instance(are_coordinates_within_distance_1, ValueError,
                            'initial coordinates are outside the unit circle (euclidean distance)')
@raise_if_false_on_instance(member_enforcer('coordinates', Iterable), AttributeError,
                            'Missing coordinates after initialization')
@raise_if_false_on_class(member_enforcer('testme', types.FunctionType), AttributeError,
                         "Checks if a method called testme is present")
@raise_if_false_on_class(member_enforcer('MULTIPLIER', float), AttributeError,
                         "Checks if a float multiplier is set on the class")
class LibraryClass(metaclass=HasEnforcedRules):
    MULTIPLIER = 0.5
    NOT_USED_MULTIPLIER = 20.0

    def __init__(self, name: str, *args):
        self.name = name
        self.coordinates = list(args)

    def testme(self):
        print(self.name)

    pass
